(* This module defines a pretty-printer for System F types and terms
   in internal form. It relies on the basic machinery provided by
   [Export]. *)

module B = Buffer
open Types
open Pprint
open Coercions
open Terms

(* ------------------------------------------------------------------------- *)

(* Record labels. *)

let pl : Atom.atom -> Identifier.identifier =
  (* This call is made just once, so that all calls to [Print] print
     record labels in a consistent way. *)
  Export.mkglobal()

let print_label label =
  Identifier.name (pl label)

let pl label =
  text (print_label label)

(* ------------------------------------------------------------------------- *)

(* Quantifiers. *)

let pq = function
  | QForall ->
      text "forall"
  | QExists ->
      text "exists"

(* ------------------------------------------------------------------------- *)

(* Term and type variables. *)

let pvar env x =
  text (Identifier.name (Export.resolve env x))

(* ------------------------------------------------------------------------- *)

(* Syntactic elements for the pretty-printer. *)

let arrow =
  text " ->" ^^ softline

let lambda =
  text "fun "

let colon =
  text " :" ^^ softline

let equal =
  text " =" ^^ softline

(* ------------------------------------------------------------------------- *)

(* Syntactic idioms for the pretty-printer. *)

(* [heading head body] prints [head]; breaks a line and indents by 2,
   if necessary; then prints [body]. *)

let heading head body =
  nest 2 (
    group head ^^ linebreak ^^
    body
  )

(* [jump body] either displays a space, followed with [body], followed
   with a space, all on a single line; or breaks a line, prints [body]
   at indentation 2, and breaks another line. *)

let jump body =
  group (nest 2 (line ^^ body) ^^ line)

(* [definition head body cont] prints [head]; prints [body], surrounded
   with spaces and, if necessary, indented by 2; prints the keyword [in];
   breaks a line, if necessary; and prints [cont]. *)

let definition head body cont =
  group (
    group head ^^ jump body ^^ text "in"
  ) ^^ line ^^
  cont

(* ------------------------------------------------------------------------- *)

(* Records. *)

let pfield label sep field =
  group (pl label ^^ sep ^^ field)

let pfields sep f fields =
  if Atom.AtomMap.is_empty fields then
    braces empty
  else
    braces (
      jump (
	sepmap
	(semi ^^ line) (* [softline] would allow several fields per line, with wrapping *)
	(fun (label, field) -> pfield label sep (f field))
	(Atom.AtomMap.to_list fields)
      )
    )

(* ------------------------------------------------------------------------- *)

(* Types. *)

(* In order to provide proper parenthesization, the structure of the
   printing functions reflects that of the parser: there are several
   levels of priority. *)

let rec pty0 env ty =
  match ty with
  | TyBoundVar _ ->
      assert false
  | TyFreeVar a ->
      pvar env a
  | TyRecord fields ->
      pfields colon (pty env) fields
  | _ ->
      parens (pty env ty)

and pty env ty =
  group (
    match ty with
    | TyArrow (domain, codomain) ->
	pty0 env domain ^^
	arrow ^^
	pty env codomain
    | TyQuantified (q, body) ->
	let a = Atom.fresh (Types.hint body) in
	let env = Export.bind env a in
	nest 2 (
	  pq q ^^ space ^^
	    pvar env a ^^ dot ^^ line ^^
	    pty env (fill body (TyFreeVar a))
	)
    | _ ->
	pty0 env ty
  )

(* ------------------------------------------------------------------------- *)

(* Coercions. *)

let forall = text "forall"

let mode b = text (if b then "<" else ">")

let rec pcoercion0 env (loc, co) = 
  match co with
  | CoId -> text "_"
  | CoIntroForall -> text "+forall" 
  | CoElimExists -> text "-exists" 
  | CoDistrib -> text "distrib"
  | CoDistribLabel label -> text "distrib" ^^ dot ^^ pl label
  | _ ->
      parens (pcoercion env (loc, co))

and pcoercion env (loc, co) =
  group (
  match co with
  | CoElimForall (m, ty) ->
      text "-forall" ^^ mode m ^^ space ^^ pty env ty

  | CoIntroExists (m, ty) ->
      text "+exists" ^^ mode m ^^ space ^^ pty env ty

  | CoQuantifier (q, a, co) ->
      let env = Export.bind env a in
      heading 
        (pq q ^^ space ^^ pvar env a ^^ dot ^^ line)
	(pcoercion env co)

  | CoTrans (co1, co2) ->
	pcoercion env co1 ^^
	semi ^^ softline ^^
	pcoercion env co2
      
  | CoArrow (co1, co2) -> 
	pcoercion env co1 ^^
	arrow ^^
	pcoercion env co2

  | _ ->
      pcoercion0 env (loc, co)
      
 )

(* ------------------------------------------------------------------------- *)

(* Terms. *)

let rec pterm0 env (loc, term) =
  match term with
  | TeVar x ->
      pvar env x
  | TeRecord fields ->
      pfields (equal ^^ softline) (pterm env) fields
  | TeProj (term, label) ->
      pterm0 env term ^^ dot ^^ pl label
  | _ ->
      parens (pterm env (loc, term))

and pterm1 env (loc, term) =
  group
    (match term with
    | TeApp (term1, term2) ->
        pterm1 env term1 ^^ line ^^ pterm0 env term2
    | TeTyApp (term, ty) ->
        pterm1 env term ^^ line ^^ brackets (pty env ty)
    | _ ->
        pterm0 env (loc, term)
    )

and pterm env (loc, term) =
  group
    (match term with
    | TeFun (x, domain, codomain, body) -> 
	let env = Export.bind env x in
	heading
	  (
	   lambda ^^ 
           parens (pvar env x ^^ colon ^^ pty env domain) ^^
           begin match codomain with
           | Some ty -> colon ^^ pty env ty ^^ equal
           | None -> equal
           end
	  )
          (pterm env body)
    | TeRecFun (f, x, domain, codomain, body) ->
	let env = Export.bind env f in
	let env = Export.bind env x in
	heading
	  (
	   lambda ^^ pvar env f ^^ line ^^
           parens (pvar env x ^^ colon ^^ pty env domain) ^^
    	   colon ^^ pty env codomain ^^ equal
	  )
          (pterm env body)
    | TeTyAbs (a, term) ->
	let env = Export.bind env a in
	heading
	  (lambda ^^ pvar env a ^^ arrow)
	  (pterm env term)
    | TeLet (x, term1, term2) ->
	let term1 = pterm env term1 in
	let env = Export.bind env x in
	definition
	  (text "let" ^^ line ^^ pvar env x ^^ equal)
	  term1
          (pterm env term2)
    | TePack (term, existential) ->
	text "pack" ^^ line ^^
	pterm env term ^^ colon ^^
	pty env (TyQuantified (QExists, existential))
    | TeUnpack (a, x, term1, term2) ->
	let term1 = pterm env term1 in
	let env = Export.bind env a in
	let env = Export.bind env x in
	definition
	  (text "unpack" ^^ line ^^ pvar env a ^^ comma ^^ line ^^ pvar env x ^^ equal)
	  term1
	  (pterm env term2)
    | TeCoerce (co, term) ->
        angles (pcoercion env co) ^^ softline ^^ pterm env term
                                                   
    | _ ->
	pterm1 env (loc, term)
    )

(* ------------------------------------------------------------------------- *)

(* Wrapping up. *)

let b =
  B.create 2048

let doc2string doc =
  B.clear b;
  Pprint.Buffer.pretty 0.95 78 b doc;
  B.contents b

let print_type env ty =
  doc2string (pty env ty)

let print_term env term =
  doc2string (pterm env term)

