open Atom

(* ------------------------------------------------------------------------- *)

(* Quantifiers. *)

type quantifier = Syntax.quantifier =
  | QForall
  | QExists

(* ------------------------------------------------------------------------- *)

(* Types. *)

type ftype =
  | TyBoundVar of int
  | TyFreeVar of atom
  | TyArrow of ftype * ftype
  | TyRecord of ftype AtomMap.t
  | TyQuantified of quantifier * ftype_context

(* Internally, a type context is just a type in which the de Bruijn index 0 is
   allowed to occur. *)

(* A type context also records an identifier, whose sole purpose is to serve as
   a hint when pretty-printing. It can otherwise be ignored. *)

and ftype_context =
    Identifier.identifier * ftype

(* ------------------------------------------------------------------------- *)

(* Well-formedness. *)

let rec wf i = function
  | TyBoundVar j ->
      j < i
  | TyFreeVar _ ->
      true
  | TyArrow (domain, codomain) ->
      wf i domain && wf i codomain
  | TyRecord fields ->
      AtomMap.fold (fun _ field accu -> accu && wf i field) fields true
  | TyQuantified (_, (_, body)) ->
      wf (i+1) body

(* ------------------------------------------------------------------------- *)

(* Equality. *)

let rec equal ty1 ty2 =
  match ty1, ty2 with
  | TyBoundVar i1, TyBoundVar i2 ->
      i1 = i2
  | TyFreeVar a1, TyFreeVar a2 ->
      Atom.equal a1 a2
  | TyArrow (domain1, codomain1), TyArrow (domain2, codomain2) ->
      equal domain1 domain2 && equal codomain1 codomain2
  | TyRecord fields1, TyRecord fields2 ->
      AtomMap.equal equal fields1 fields2
  | TyQuantified (QForall, (_, body1)), TyQuantified (QForall, (_, body2))
  | TyQuantified (QExists, (_, body1)), TyQuantified (QExists, (_, body2)) ->
      equal body1 body2
  | _, _ ->
      false

(* ------------------------------------------------------------------------- *)

(* Occurs check. *)

exception Occurs

let rec occurs a ty =
  match ty with
  | TyFreeVar b when Atom.equal a b ->
      raise Occurs
  | TyFreeVar _
  | TyBoundVar _ ->
      ()
  | TyArrow (domain, codomain) ->
      occurs a domain;
      occurs a codomain
  | TyRecord fields ->
      AtomMap.iter (fun _ field -> occurs a field) fields
  | TyQuantified (q, (_, body)) ->
      occurs a body

let occurs a ty =
  try
    occurs a ty;
    false
  with Occurs ->
    true

(* ------------------------------------------------------------------------- *)

(* Context creation. *)

let rec abstract a i ty =
  match ty with
  | TyFreeVar b when Atom.equal a b ->
      TyBoundVar i
  | TyFreeVar _
  | TyBoundVar _ ->
      ty
  | TyArrow (domain, codomain) ->
      TyArrow (abstract a i domain, abstract a i codomain)
  | TyRecord fields ->
      TyRecord (AtomMap.map (abstract a i) fields)
  | TyQuantified (q, (hint, body)) ->
      TyQuantified (q, (hint, abstract a (i+1) body))

let abstract a ty =
  let hint = Atom.identifier a in
  hint, abstract a 0 ty

(* ------------------------------------------------------------------------- *)

(* Context elimination. *)

let rec fill i ty c =
  match c with
  | TyBoundVar j when i = j ->
      ty
  | TyBoundVar _
  | TyFreeVar _ ->
      c
  | TyArrow (domain, codomain) ->
      TyArrow (fill i ty domain, fill i ty codomain)
  | TyRecord fields ->
      TyRecord (AtomMap.map (fill i ty) fields)
  | TyQuantified (q, (hint, body)) ->
      TyQuantified (q, (hint, fill (i+1) ty body))

let fill (_, c) ty =
  fill 0 ty c

(* ------------------------------------------------------------------------- *)

(* Hints. *)

let hint (hint, _) =
  hint

(* ------------------------------------------------------------------------- *)

(* Matching. *)

let matching cty ty2 =

  (* Create an atom to stand for the flexible variable. Instantiate the context
     with it. *)

  let flex = Atom.fresh (hint cty) in
  let ty1 = fill cty (TyFreeVar flex) in

  (* Create a reference that holds a candidate value for the flexible variable. *)

  let candidate = ref None in

  (* The goal is now to check the types [ty1] and [ty2] for equality. The type
     [ty1] contains occurrences of the flexible variable [flex]; we must check
     that these are equated with well-formed types, and that there is at most
     one such type. *)

   let rec equal ty1 ty2 =
    match ty1, ty2 with
    | TyFreeVar a, _ when Atom.equal a flex ->
	
	(* The left-hand side is the flexible variable. *)

	(* If we did not yet have a candidate, [ty2] becomes a candidate,
	   provided it is a well-formed type, i.e. that no variable escapes
	   its scope. If we did have one, check that [ty2] is equal to
	   it. This is a real equality check (no matching involved), since
	   neither [ty1] nor [ty2] contain occurrences of the flexible
	   variable. *)

	begin match !candidate with
	| None ->
	    wf 0 ty2 && (candidate := Some ty2; true)
	| Some ty1 ->
	    equal ty1 ty2
	end

    | TyBoundVar i1, TyBoundVar i2 ->
	i1 = i2
    | TyFreeVar a1, TyFreeVar a2 ->
	Atom.equal a1 a2
    | TyArrow (domain1, codomain1), TyArrow (domain2, codomain2) ->
	equal domain1 domain2 && equal codomain1 codomain2
    | TyRecord fields1, TyRecord fields2 ->
	AtomMap.equal equal fields1 fields2
    | TyQuantified (QForall, (_, body1)), TyQuantified (QForall, (_, body2))
    | TyQuantified (QExists, (_, body1)), TyQuantified (QExists, (_, body2)) ->
	equal body1 body2
    | _, _ ->
	false
  in
  if equal ty1 ty2 then Some !candidate
  else None

let matches x ty2 = matching x ty2 <> None

(* ------------------------------------------------------------------------- *)

(* Typing environments map term variables to types. *)

type tenv =
    ftype AtomMap.t

let empty : tenv =
  AtomMap.empty

let lookup : atom -> tenv -> ftype =
  fun x tenv ->
    try
      AtomMap.find x tenv
    with Not_found ->
      assert false (* should not happen, if there is a binding of [x] in [tenv] *)

let bind : atom -> ftype -> tenv -> tenv =
  AtomMap.add

